/*  
   TAF 0.1 - TAF Ain't Fortune - An expansion on the Fortune utility. 
   Copyright (C) 1998 Michael Pearson

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

   I can be contacted at alcaron@ozemail.com.au. Feel free to mail me any
   diffs, bug reports, comments, and suchlike
 */

/* Includes */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <ctype.h>
#include "config.h"

/* Prototypes */
int GetFileFromDisk (char filename[], char *list[], int *numentries);
void ParseList (char *list[], char *output[], int length, int ssize, int line, int secnum, int numentries);
char *GetLineFromList (char in[], int length);
void ShowMeString (char in[]);
void DoOutput (char *output[], int numlines, int newline);
int ParseArgs (int numargs, char *arg[], char **filename, int *length, int *ssize, int *line, int *newline, int *secnum, int numentries);
int IsNumber (char in[]);
void DumpVersion ();
void DumpHelp ();
int SanityChecks (int *length, int *ssize, int *line, int *secnum, int *numentries);
extern char *strdup(char *); 

int
main (int argvc, char *argv[])
{
  /* Defaults */
  char *filename = FILENAME;	/* File to get from disk and put in
				   list                       */
  int length = LENGTH;		/* Length of string to output on exit */
  int ssize = SSIZE;		/* Size of each section       */
  int line = LINE;		/* Line of section to output, 0 for
				   all */
  int newline = NEWLINE;	/* Do we put a newline on the end of each output? */
  int secnum = SECNUM;		/* Section number to get, 0 for Rand */

  /* Other variables */
  int numentries;		/* Number of entries in list  */
  int numlines;			/* Number of lines to output  */

  char *list[MAX];		/* List which stores the file */
  char *output[MAX];		/* Stuff to dump to screen    */
  char **ptr;			/* Because of the way C handles passing pointers we need this */

  /* Set up screwy filename pointer */
  ptr = &filename;

  /* If there are command line args, parse them */
  if (argvc > 1)
    switch (ParseArgs (argvc, argv, ptr, &length, &ssize, &line, &newline, &secnum, numentries))
      {
      case -1:
	return -1;
      case 1:
	return 0;
      }

  /* Does the file exist? If so, continue */
  if (GetFileFromDisk (filename, list, &numentries) != 0)
    {
      fprintf (stderr, "ERROR: Could not load %s from disk! Exiting\n", filename);
      return 1;
    }
  else
    {
      /* If we want all the lines, number of lines is the size of each sec */
      if (line == -1)
	numlines = ssize;
      else
	numlines = 0;

      if (!SanityChecks (&length, &ssize, &line, &secnum, &numentries))
	return -1;

      ParseList (list, output, length, ssize, line, secnum, numentries);
      DoOutput (output, numlines, newline);
      return 0;
    }
}

/* Grab file from disk, and put it in the list */
/* Called From: main                           */
int
GetFileFromDisk (char filename[], char *list[], int *numentries)
{
  FILE *fp;
  char gotstuff[256];
  int count = 0;

  /* Does the file exist? */
  if ((fp = fopen (filename, "r")) == NULL)
    {
      return 1;
    }
  else
    {
      /* Get stuff from file until EOF or MAX is hit */
      while ((fgets (gotstuff, 256, fp) != NULL) && count <= MAX)
	{
	  gotstuff[strlen (gotstuff) - 1] = ' ';
	  if ((list[count] = (char *) malloc (strlen (gotstuff) + 1)) == NULL)
	    {
	      fprintf (stderr, "ERROR: Memory allocation error!\n");
	      exit (-1);
	    }
	  strcpy (list[count], gotstuff);
	  count++;
	}
      *numentries = count;
      return 0;
    }
}

/* Get the lines from the list, and put them in output */
/* Called From: main                                   */
void
ParseList (char *list[], char *output[], int length, int ssize, int line, int secnum, int numentries)
{
  int count = 0;
  int sstart;


  /* If the sector number to show is lower than zero, use a random one */
  if (secnum == -1)
    {
      srand (time (0));
      secnum = numentries / (ssize + 1) * (float) rand () / RAND_MAX;
    }

  /* Sstart will be where the sector starts from */
  sstart = secnum * (ssize + 1);

  /* If line is lower than zero, do all the lines. If not, show the right line */
  if (line == -1)
    {
      /* Do all the lines */
      for (count = 0; count <= ssize; count++)
	{
	  output[count] = GetLineFromList (list[count + sstart], length);
	}
    }
  else
    {
      /* Do just the one  */
      output[0] = GetLineFromList (list[sstart + line], length);

    }
}

/* Does the equalivant of strdup'ing the string,  */
/* but also adds or subtracts bits from it accor  */
/* to length
   Called from: ParseList                         */
char
 *
GetLineFromList (char *in, int length)
{
  char *ptr;
  int count;
  char whitespace[256];

  /* If length is lower than zero set it to length of string */
  if (length == -1)
    length = strlen (in);

  /* Makes sure there's enough space for the string, and allocates it */
  if ((ptr = (char *) malloc (length + 1)) == NULL)
    {
      fprintf (stderr, "ERROR: Memory allocation error!");
      exit (-1);
    }

  /* If string is shorter than length, add whitespaces, if longer, crop string */
  if (strlen (in) < length)
    {
      /* Make the whitespace */
      for (count = 0; count <= (length - strlen (in) - 1); count++)
	whitespace[count] = ' ';
      whitespace[count] = '\0';

      /* Insert whitespace */
      strcpy (ptr, in);
      strcat (ptr, whitespace);
    }
  else
    {
      strncpy (ptr, in, length);
    }
  return ptr;
}

/* This is a debugging function to show me the contents of a string in numerical format */
/* Called by: GetLineFromList                                                           */
void
ShowMeString (char in[])
{
  int count;
  printf ("String: %s\nString Contents: ", in);
  for (count = 0; count <= strlen (in); count++)
    {
      printf ("%d, ", in[count]);
    }
  printf ("\n");
}

/* Dump all the lines to the screen 
   Called by: main                  */
void
DoOutput (char *output[], int numlines, int newline)
{
  int count;
  for (count = 0; count <= numlines; count++)
    {
      printf ("%s", output[count]);
      if (newline)
	printf ("\n");
    }
}

/* Parse the command line arguments
   Called by: main                  */
/* FIXME: I really should unspaghetti code this someday. 
   Mail me a diff at alcaron@ozemail.com.au if you do. */
int
ParseArgs (int numargs, char *arg[], char **filename, int *length, int *ssize, int *line, int *newline, int *secnum, int numentries)
{
  int count;
  int done_length = 0;
  int done_ssize = 0;
  int done_line = 0;
  int done_newline = 0;
  int done_secnum = 0;
  int done_filename = 0;

  for (count = 1; count < numargs; count++)
    {
      /* switch wont work for strings, so we'll use a set of else ifs here */
      /* Help */
      if (strcmp (arg[count], "-h") == 0 || strcmp (arg[count], "--help") == 0)
	{
	  DumpHelp ();
	  return (1);
	}

      /* Version */
      if (strcmp (arg[count], "-v") == 0 || strcmp (arg[count], "--version") == 0)
	{
	  DumpVersion ();
	  return (1);
	}

      /* Newline */
      if (strcmp (arg[count], "-n") == 0 || strcmp (arg[count], "--nonewline") == 0)
	{
	  if (done_newline)
	    {
	      fprintf (stderr, "WARNING: Newline already defined on argument %d, redefining.\n", done_newline);
	    }
	  done_newline = count;
	  *newline = 0;
	}

      /* Check if we have another argument */
      else if (count == numargs - 1)
	{
	  fprintf (stderr, "ERROR: No argument for %s or argument %s not recognised.\nERROR: Type %s --help for information about arguments.\n", arg[count], arg[count], arg[0]);
	  return (-1);
	}

      /* Filename */
      else if (strcmp (arg[count], "-f") == 0 || strcmp (arg[count], "--filename") == 0)
	{
	  if (done_filename)
	    {
	      fprintf (stderr, "WARNING: Filename already defined on argument %d, redefining.\n", done_filename);
	    }

	  if ((*filename = (char *) strdup(arg[count + 1])) == NULL)
	    {
	      fprintf (stderr, "ERROR: Memory allocation error.\n");
	      return (-1);
	    }
	  done_filename = count;
	  count++;
	}

      /* Length */
      else if (strcmp (arg[count], "-l") == 0 || strcmp (arg[count], "--length") == 0)
	{
	  if (done_length)
	    {
	      fprintf (stderr, "WARNING: Length already defined on argument %d, redefining.\n", done_length);
	    }

	  if (!IsNumber (arg[count + 1]))
	    {
	      fprintf (stderr, "ERROR: %s isn't a number. A number is required for argument %s.\n", arg[count + 1], arg[count]);
	      return (-1);
	    }
	  done_length = count;
	  *length = atoi (arg[count + 1]);
	  count++;
	}

      /* Ssize */
      else if (strcmp (arg[count], "-s") == 0 || strcmp (arg[count], "--ssize") == 0)
	{
	  if (done_ssize)
	    {
	      fprintf (stderr, "WARNING: Sector Size already defined on argument %d, redefining.\n", done_ssize);
	    }

	  if (!IsNumber (arg[count + 1]))
	    {
	      fprintf (stderr, "ERROR: %s isn't a number. A number is required for argument %s.\n", arg[count + 1], arg[count]);
	      return (-1);
	    }
	  done_ssize = count;
	  *ssize = atoi (arg[count + 1]) - 1;
	  count++;
	}

      /* Line */
      else if (strcmp (arg[count], "-i") == 0 || strcmp (arg[count], "--line") == 0)
	{
	  if (done_line)
	    {
	      fprintf (stderr, "WARNING: Line number already defined on argument %d, redefining.\n", done_line);
	    }

	  if (!IsNumber (arg[count + 1]))
	    {
	      fprintf (stderr, "ERROR: %s isn't a number. A number is required for argument %s.\n", arg[count + 1], arg[count]);
	      return (-1);
	    }
	  done_line = count;
	  *line = atoi (arg[count + 1]) - 1;
	  count++;
	}

      /* Secnum */
      else if (strcmp (arg[count], "-e") == 0 || strcmp (arg[count], "--secnum") == 0)
	{
	  if (done_secnum)
	    {
	      fprintf (stderr, "WARNING: Sector number already defined on argument %d, redefining.\n", done_line);
	    }

	  if (!IsNumber (arg[count + 1]))
	    {
	      fprintf (stderr, "ERROR: %s isn't a number. A positive and whole number is required for argument %s.\n", arg[count + 1], arg[count]);
	      return (-1);
	    }
	  done_secnum = count;
	  *secnum = atoi (arg[count + 1]) - 1;
	  count++;
	}

      /* Unrecognized */
      else
	{
	  fprintf (stderr, "ERROR: Argument %s not recognized. Type %s --help for a list of valid arguments\n", arg[1], arg[0]);
	  exit (-1);
	}
    }
  
  /* Useless return here - but it kills the compiler warnings */
  fprintf(stderr,"ERROR: Hit the useless return while parsing args.\n");
  fprintf(stderr,"       That really shouldn't have happened. Mail\n");
  fprintf(stderr,"       alcaron@ozemail.com.au with -exactly- what\n");
  fprintf(stderr,"       you did, and what version of gcc you're using.\n\n");
  return (-1);

  /* Phew! */
}

/* Dump the version number of the program, my name, 
   copyright info, and where I can be contacted at.
   Called by: ParseArgs */
void
DumpVersion ()
{
  printf ("%s\n", LONG_VERSION);
  printf ("Copyright (C) 1998 Michael Pearson\n");
  printf ("\n");
  printf ("%s comes with NO WARRANTY, either expressed or implied,\n", VERSION);
  printf ("and is distributed under the GNU Public License. You are free\n");
  printf ("to modify, distribute, and use %s under the terms and\n", VERSION);
  printf ("conditions in the file COPYING.\n.");
  printf ("\n");
  printf ("If the file COPYING did not come with this program please mail\n");
  printf ("me at alcaron@ozemail.com.au and inform me who you received the\n");
  printf ("program from. Also view the GNU Public License at:\n");
  printf ("\n");
  printf ("http://www.gnu.org/copyleft/gpl.html\n");
  printf ("\n");
  printf ("%s is a program which reads a file and outputs a random or\n", VERSION);
  printf ("specified line, depending on the command line options. You \n");
  printf ("can get more information about the options by typing --help\n");
  printf ("after the command name.\n\n");
}

/* Show help */
void
DumpHelp ()
{
  printf ("%s\n", LONG_VERSION);
  printf ("Copyright (C) 1998 Michael Pearson\n");
  printf ("\n");
  printf ("-h or --help :           This help screen\n");
  printf ("-v or --version :        Program version\n");
  printf ("-f or --filename :       The file to load into the parser\n");
  printf ("-l or --length :	    The length of the string to output, 0 for\n");
  printf ("                         the original length of the string\n");
  printf ("-s or --ssize :          The size of each sector\n");
  printf ("-n or --nonewline :      Don't put a newline at the end of each output\n");
  printf ("-i or --line :           Output this line of the sector, -1 for all\n");
  printf ("-e or --secnum :         Output this sector, -1 for random\n");
  printf ("\n");
  printf ("Each file is divided up into sectors of ssize, and by default a random\n");
  printf ("sector is chosen from these. If no line number is specified TAF will\n");
  printf ("show all the lines of a sector.\n\n");
}

/* A procedure to check each bit of a string and make sure it's all digits
   Called by:  ParseArgs */
int
IsNumber (char in[])
{
  int count;
  for (count = 0; count <= (strlen (in) - 1); count++)
    {
      if (!isdigit(in[count]) && in[count] != '-')
	return 0;
    }
  return 1;
}

/* This little gem makes sure the program doesn't segfault because the user enter bad info
   Called by: Main */
int
SanityChecks (int *length, int *ssize, int *line, int *secnum, int *numentries)
{
  /* First check to see that we don't have any numbers below -1 
     and if we do, default them to -1 */
  if (*length < -1)
    {
      fprintf (stderr, "WARNING: Length is below -1, defaulting.\n");
      *length = -1;
    }
  if (*secnum < -1)
    {
      fprintf (stderr, "WARNING: Sector Number is below 0, defaulting to random.\n");
      *secnum = -1;
    }

  if (*line < -1)
    {
      fprintf (stderr, "WARNING: Line Number is below 0, defaulting to random.\n");
      *line = -1;
    }

  /* Check to see that sector size is actually a positive, and there are lines
     in the infile */
  if (*ssize < 0)
    {
      fprintf (stderr, "ERROR: Sector size is below 1.\n");
      return 0;
    }
  if (*numentries < 1)
    {
      fprintf (stderr, "ERROR: Infile contains no data.\n");
      return 0;
    }

  /* Check to see if sector number multiplied by sector size is more than
     the capacity of the file */
  if (*secnum != -1 && ((*secnum + 1) * (*ssize + 1)) > *numentries)
    {
      fprintf (stderr, "ERROR: Sector number multiplied by sector size exceeds file length.\n");
      return 0;
    }

  /* Check to see if line number is more than sector size */
  if (*line != -1 && *ssize < *line)
    {
      fprintf (stderr, "ERROR: Line number exceeds sector size.\n");
      return 0;
    }

  /* Check to see if sector size is more than size of file */
  if (*ssize > *numentries)
    {
      fprintf (stderr, "ERROR: Sector size exceeds file length.\n");
      return 0;
    }

  /* Sanity checks passed, return true */
  return 1;
}

/* Inertia Creeps. */
